/* ------------------------------------------------------------------------
	nodes.c

	A Citadel utility to generate network message statistics.

	Written by Audin Malmin.

	With other code by various Citadel authors and Golden Spud.

	This code is Public Domain.  So do with it as you please.

	However, if you sell it then you will become very evil and I shall
	probably never speak to you again.

	There is, of course, no warranty what so ever.  This code works fine
	on my system and the few others it's been used on.  But if it makes
	your system melt down in front of your eyes then its your own fault.

	If you have questions or comments, I can be reached at:

	Magrathea @ Skaro, Olympia [skr.olm]  (CitNet)

	amalmin@halcyon.com  (Internet)

	or, if all else fails:

	Audin Malmin
	1317 Horne St. N.E.
	Olympia, WA 98516-5613

 --------------------------------------------------------------------------
 History:

 95dec14/ANM	Played with GetStr().  Took runtime from 320sec to 280sec.
				Zapped a few feof calls in readmessage...time to 276 sec.
				Created usage().
				Decreased the number of deansi() calls...saved 3 seconds.wow
				Added Spud's hash stuff.  a user hash alone saves 47seconds.
				Using hashing on roomnames dropped the time to 195 seconds!
				So we've shaved 125 seconds off the time today, i'm impressed.
 95dec16/ANM	Stuck in a bunch of Golden Spud's linked-list code.  nice
	|			and fast.
	|			Fixed a bug in stripansi().  pain in the butt, that was.
	|			Added a stricmp check after the hash check...just because...
	|			Added sort stuff.  Thought there must be a better way...
 95dec18/ANM	Version up to 5.0.
 95dec19/ANM	Added config file stuff.
				Added multiple status displays...
				And added the exclusion code...although it's not ideal yet.
 95dec25/ANM	Expanded config file options.  Got rid of the CLI stuff.
 95dec26/ANM	And even more config file options...
 95dec30/ANM	Added compressed message detection.
				Added compressed message handling...
				Added local message detection.  (excluded from list.)
				Number of Local messages are listed in nodes.lst file.
				Added attribute byte (1st and 2nd) handling and room number
				byte (1st and 2nd as well) handling...although its not used.
				Version 6.0.
 96jan02/ANM	Added network connections list.

 -------------------------------------------------------------------------- */
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <dir.h>
#include  <dos.h>
#include  <conio.h>
#include  <io.h>
#include  <alloc.h>
#include  <time.h>
#include  <ctype.h>
#include  "../stat/pragmas.h"

#define FALSE 0
#define TRUE 1
#define	LABELSIZE	31
							/* these are message attributes. */
#define COMPRESSED 0x08     // Compressed message.
#define MORE 0x40           // message has a second attribute byte.
#define BIGROOM 0x20        // Message uses 16 bits for the room number
#define LOCAL 0x01          // (attr2) message is local-only.

typedef char label[LABELSIZE];

FILE    *msgdat, *nodesout, *roomsout, *usersout,
		*nodeslst, *linkslst;

#ifdef DEBUG
FILE	*debugout;
#endif

label	version			 = "6.0";
char    inputmail[257]   = "c:\\cit\\messages\\msg.dat";
char	outputnode[257]  = "nodes.lst";
char	outputroom[257]  = "rooms.lst";
char    outputuser[257]  = "users.lst";
char	outlinklist[257] = "links.lst";
char    outnodelist[257] = "nodes.txt";

#ifdef DEBUG
char	debugfile[257]	 = "debug.out";
#endif

label	homenode	= "";
label   homeadd     = "";
label	homesoft	= "Citadel+/065";
label	homeregion  = "";
label	homestate	= "";
label	homecountry = "";
label	homephone	= "";

char    path[513];
char	address[257];
label 	soft;
char    messagetext[8193];
label   nodename;
label	author;
label	roomname;
label   region;
label	state;
label   ncountry;
label   phone;
label	datestr;
label	alias1;
label	alias2;
char	sig[81];
char	softname[257];

unsigned char attr = 0,
			  attr2 = 0;

int		jaydee = 0;
int		excroom = 0;
int		debug = 0;
int		listtype = 0;
int		status = 0;

static int column = 0;
const int maxcol = 78;

long	nmessages = 0;
long	messagek = 0;
int		messagebyte = 0;
int		length = 0;
int		local = 0;

time_t	starttime, endtime;

long	firsttime = 999999999L, lasttime = 0;

int		numnode = 0;
int		numroom = 0;
int		numuser = 0;
int		numlink = 0;

typedef struct	node {
		label nodename;
		label address;
		label phone;
		label region;
	    label state;
		label country;
		int software;
		int hash;
		unsigned excl:1;
		long messages;
		unsigned long msgk;
		struct node* next;
		} node;

node*	firstnode = NULL;
node*	currnode  = NULL;

typedef struct	room {
		label roomname;
		int hash;
		unsigned excl:1;
		long messages;
		unsigned long msgk;
		struct room* next;
		} room;

room*	firstroom = NULL;
room*	currroom  = NULL;

typedef struct	user {
		label username;
		int	hash;
		unsigned excl:1;
		long messages;
		unsigned long msgk;
		struct user* next;
		} user;

user*	firstuser = NULL;
user*	curruser  = NULL;

typedef struct link {
	label add1;
	label add2;
	int hash1;
	int hash2;
	unsigned long messages;
	unsigned long msgk;
	struct link* next;
} link;

link*	firstlink = NULL;
link*	currlink = NULL;

typedef struct xuser {
	char name[31];
	struct xuser* next;
} xuser;

xuser* firstxuser = NULL;
xuser* currxuser = NULL;

static	char *software[] = {
	"Autoposter",				/* 0 */
	"Gremlin's Citadel/064.2",	/* 1 */
	"Gremlin's Citadel/064",	/* 2 */
	"Gremlin's Citadel",		/* 3 */
	"Citadel+/066",			/* 4 */
	"Citadel+/066",				/* 5 */
	"Citadel+/065",				/* 6 */
	"Citadel+",					/* 7 */
	"MavenCit",					/* 8 */
	"Co-Op Citadel",			/* 9 */
	"Frog's Citadel",			/* 10 */
	"Poster",					/* 11 */
	"[no software field]"		/* 12 */
	};

static	char *softkey[] = {
	"APost",					/* 0  */
	"G/642",					/* 1  */
	"G/064",					/* 2  */
	"GCit",						/* 3  */
	"+/66",					/* 4  */
	"C+/66",					/* 5  */
	"C+/65",					/* 6  */
	"Cit+",						/* 7  */
	"Mcit",						/* 8  */
	"Co-Op",					/* 9  */
	"Frog",						/* 10 */
	"Postr",					/* 11 */
	"????"						/* 12 */
	};

int  	main(int argc, char *argv[]);
void 	startup(void);
int  	readMessage(void);
void 	nodestat(void);
void 	roomstat(void);
void    userstat(void);
void	pathstat(void);
void	linkstat(void);
int  	GetStr(FILE *fil, char *str, int mlen);
int  	LenStr(FILE *fil);
int		cLength(void);
void	error(void);
int		isaddress(const char *str);
int		Nsort(node* this, node* that);
/* could someone who understands this fix it so it isn't so redundant? */
/***********************************************************************/
node*	Nlmerge(node* p, node* q);
node*	Nlsort (node* p);
int		Usort(user* this, user* that);
user*	Ulmerge(user* p, user* q);
user*	Ulsort (user* p);
int		Rsort(room* this, room* that);
room*	Rlmerge(room* p, room* q);
room*	Rlsort (room* p);
int		Lsort(link* this, link* that);
link*	Llmerge(link* p, link* q);
link*	Llsort (link* p);
/************************************************************************/
int		hash(const char* s);
void	stripansi(char *str);
void	usage(void);
void	output(void);
void	config(void);
FILE	*pfopen(const char* name, const char* mode, const char* dirs);

/****************************************************************************/
/* main() -- Look Golden Spud!  it returns an int! :-)                      */
/****************************************************************************/
int main(int argc, char *argv[])
{
clock_t ticks;

	 strcpy(softname, argv[0]);
     strlwr(softname);

     printf("\n"
			"Nodelist %s (%s %s)\n"
			"\n"
			"By Magrathea, Golden Spud, and Mystery Man\n"
			"\n"
			"   Magrathea @ Skaro, Olympia, (360)491-9605\n"
			"   Golden Spud @ Golden Spud's Greashouse\n"
			"   Mystery Man @ Higher Ground\n"
			"\n"
			"With other code by Brent Bottles and various other Citadel authors.\n"
			"\n" , version, __DATE__, __TIME__);

	config();

	printf("Input filename..: %s\n"
		"Node stat output: %-20s  Node list output: %s\n"
		"Room stat output: %-20s  User stat output: %s\n"
		"Home nodename...: %-20s  Home address....: %s\n"
		"Home software...: %-20s  Home region.....: %s\n"
		"Home state......: %-20s  Home country....: %s\n"
		"Home phone......: %s\n"
		"Status type.....: %d\n"
		"Exclude Jaydee?.: %-20s  Debug mode......: %s\n"
		"\n"
		"New Node:   New Room:   New User:   New Link: !\n\n",
		inputmail, outputnode, outnodelist, outputroom,
		outputuser, homenode, homeadd, homesoft, homeregion, homestate,
		homecountry, homephone, status, (!jaydee) ? "Yes" : "No",
		(debug) ? "On" : "Off");

	startup();

	while(readMessage())		// Main loop...
		messagebyte = 0;

	output();

	fcloseall();

	ticks = clock();
	printf("Elapsed time...Ticks: %ld, Seconds: %f\n", ticks, ((float)ticks/CLK_TCK));

	return(0);
}

/*
 * Read a message...then calls the various stat functions...
 */
int readMessage(void)
{
int c, c2;

	nodename[0]		= 0;
	sig[0]			= 0;
	path[0]			= 0;
	address[0]		= 0;
	soft[0]			= 0;
	roomname[0]		= 0;
	author[0]		= 0;
	region[0]		= 0;
	state[0]		= 0;
	phone[0]		= 0;
	ncountry[0]		= 0;
	messagetext[0]	= 0;
	attr			= 0;
	attr2			= 0;

	/* find start of message */
	do {
		c = fgetc(msgdat);
	} while (c != 0xFF && c != EOF);

	if (c == EOF)
		return 0;

	attr = fgetc(msgdat);		// Get the first attribute byte.

	if((attr & MORE) != 0)		// Get second attribute byte, if there is one.
		attr2 = fgetc(msgdat);

	if((attr & BIGROOM) != 0)	// room is number 255 or greater, thus there
	{							// is 16 bits of room number.
		(void)fgetc(msgdat);
		(void)fgetc(msgdat);
	}
	else						// room is number 254 or less, so theres only
		(void)fgetc(msgdat);	// 8 bits of room number.

	do {
		c = fgetc(msgdat);

		switch (c)
		{
			case 'A':	(void)GetStr(msgdat, author,   30);	break;
			case 'D':	(void)GetStr(msgdat, datestr,  30);	break;
			case 'R':	(void)GetStr(msgdat, roomname, 30);	break;
			case 's':	(void)GetStr(msgdat, soft,     30);	break;
			case 'P':	(void)GetStr(msgdat, path,    256);	break;
			case 'O':	(void)GetStr(msgdat, nodename, 30);	break;
			case 'o':	(void)GetStr(msgdat, region,   30);	break;
			case 'Q':	(void)GetStr(msgdat, state,    30);	break;
			case 'U':	(void)GetStr(msgdat, ncountry, 30);	break;
			case 'H':	(void)GetStr(msgdat, phone,    30);	break;
			case '.':	(void)GetStr(msgdat, sig,      80);	break;
			case '_':	messagebyte += LenStr(msgdat);		break;
			case 'n':	messagebyte += LenStr(msgdat);		break;
			case 'N':	messagebyte += LenStr(msgdat);		break;
			case '%':	messagebyte += LenStr(msgdat);		break;
			case 'M':										break;
			default:
				do {
					c2 = fgetc(msgdat);
				} while(c2 && c2 !=EOF);
				break;
		}
	} while (c != 'M' && c != EOF);

	if (c == EOF)
		return 0;

	length = GetStr(msgdat, messagetext, 8192);

	if((attr2 & LOCAL) != 0)  // it's a local message, skip it.
	{
		local++;
		return 1;
	}

	if((attr & COMPRESSED) != 0)	// is the message compressed?
		messagebyte += cLength();	// Yes...send it through cLength.
	else
		messagebyte += length;		// No...

	if(!nodename[0] && sig[0])	// Message comes from the local node.
	{							// So we have to give it network fields...
		strcpy(nodename, homenode);
		strcpy(path, homeadd+1);
		strcpy(soft, homesoft);
		strcpy(region, homeregion);
		strcpy(state, homestate);
		strcpy(ncountry, homecountry);
		strcpy(phone, homephone);
	}

	if(nodename[0])		// Message is networked if it has a nodename.
	{
		nmessages++;
		messagek += messagebyte;
		nodestat();
		roomstat();
		userstat();
		pathstat();
	}
	return 1;
}

/*******************************************************/
/* nodestat() -- records node statistical information. */
/*******************************************************/
void nodestat(void)
{
int nodenum, lpc, loop, h;
node* lp;
node* newnode;
char *bang;

	strcpy(address, path);

	bang = strchr(address, '!');

	if(bang)
		*bang = 0;

	if(!isaddress(address))
		address[0] = 0;

	if(strchr(nodename, ''))
		stripansi(nodename);

	h = hash(nodename);

	for (lpc = 0, lp = firstnode, nodenum = -1; lpc < numnode; ++lpc, lp = lp->next)
		if(h == lp->hash)
			if(stricmp(nodename, lp->nodename) == 0) {
				nodenum = lpc;
				break;
				}

	if (nodenum==-1)
	{
		numnode++;
		nodenum=numnode-1;

		if(currnode == NULL) {
			/* no current nodes. */
			firstnode = currnode = (node *)farmalloc(sizeof(node));
			currnode->next = NULL;
			}
		else {
			/* list has at least one entry. */
			newnode = (node *)farmalloc(sizeof(node));
			newnode->next = NULL;
			currnode->next = newnode;
			currnode = newnode;
			}

		if(debug) {
			printf("Newnode:Coreleft        : %u\n", coreleft());
			printf("Newnode:FarCoreleft     : %lu\n", farcoreleft());
		}

		strcpy(currnode->nodename, nodename);
		strcpy(currnode->address, address);
		strcpy(currnode->region, region);
		strcpy(currnode->state, state);
		strcpy(currnode->country, ncountry);
		strcpy(currnode->phone, phone);

		for(loop=0; loop<12; loop++)
			if(strstr(soft, software[loop]))
				break;

		currnode->software = loop;
		currnode->hash = h;
		currnode->messages = 1;
		currnode->msgk = messagebyte;

		/* exclude unwanted nodes */
		for (currxuser = firstxuser; currxuser != NULL; currxuser = currxuser->next)
			if (stricmp(currnode->nodename, currxuser->name) == 0) {
				if(!status)
					fputc('', stdout);
				currnode->excl = 1;
			}
			else
				currnode->excl = 0;

		if(debug)
			printf("new node: [%30s] (%7s), %s\n", nodename, address, soft);
		else {
			if(!status)
				fputc('', stdout);
			else
				if (status == 1) {
					printf("\rNode:%3d. %-30s", numnode, nodename);
				} else if (status == 2) {
					if (column + 33 > maxcol) {
						printf("\n");
						column = 0;
						}
					column += printf("N:%-30s  ", nodename);
					}
		}
	} else
	{
		if(address[0] && stricmp(lp->address, address))
		{
			if(debug)
				printf(" address: [%30s] %s to %s\n",
						nodename, lp->address, address);

			strcpy(lp->address, address);
		}

		if(soft[0] && strstr(soft, software[lp->software]) == 0)
		{
			for(loop=0; loop<12; loop++)
				if(strstr(soft, software[loop]))
					break;

			if(debug)
				printf("software: [%30s] %s to %s\n",
				nodename, softkey[lp->software], softkey[loop]);

			lp->software = loop;
		}
	lp->messages++;
	lp->msgk += messagebyte;
	}
}

/******************************************************/
/* roomstat() -- records room statistical information */
/******************************************************/
void roomstat(void)
{
int roomnum, h, lpc;
room* lp;
room* newroom;

	if(strchr(roomname, ''))
		stripansi(roomname);

	h = hash(roomname);

	for(lpc = 0, lp = firstroom, roomnum = -1; lpc < numroom; ++lpc, lp = lp->next)
		if(h == lp->hash)
			if(stricmp(roomname, lp->roomname) == 0) {
				roomnum = lpc;
				break;
				}

	if (roomnum==-1)
	{
		numroom++;
		roomnum=numroom-1;

		if(currroom == NULL) {
			/* no elements at the moment. */
			firstroom = currroom = (room *)farmalloc(sizeof(room));
			currroom->next = NULL;
		} else {
			/* list has one or more elements */
			newroom = (room *)farmalloc(sizeof(room));
			newroom->next = NULL;
			currroom->next = newroom;
			currroom = newroom;
		}

		if(debug) {
			printf("Newroom:Coreleft        : %u\n", coreleft());
			printf("Newroom:FarCoreleft     : %lu\n", farcoreleft());
		}

		if(debug)
			if(heapcheck() != 2)
				{
				printf("The heap is fucked...exiting\n");
				printf("Coreleft   : %u %lu\n", coreleft(), coreleft());
				printf("Farcoreleft: %lu\n", farcoreleft());
				printf("You should probably write to Magrathea @ Skaro about this..\n");
				exit(1);
				}

		strcpy(currroom->roomname, roomname);
		currroom->hash = h;
		currroom->msgk = messagebyte;
		currroom->messages = 1;

		/* exclude unwanted rooms */
		for (currxuser = firstxuser; currxuser != NULL; currxuser = currxuser->next)
			if (stricmp(currroom->roomname, currxuser->name) == 0) {
				if(!status)
					fputc('', stdout);
				currroom->excl = 1;
			}
			else
				currroom->excl = 0;

		if(debug)
			printf("New Room: {%30s}\n", roomname);
		else
			if(!status)
				fputc('', stdout);
			else {
				if (status == 1) {
					printf("\rRoom:%3d. %-30s", numroom, roomname);
				} else if (status == 2) {
					if (column + 33 > maxcol) {
						printf("\n");
						column = 0;
						}
					column += printf("R:%-30s  ", roomname);
					}
			}
	} else {
		lp->msgk += messagebyte;
		lp->messages++;
	}
}

/************************************************************************/
/* userstat() -- records user info.                                     */
/************************************************************************/
void userstat(void)
{
int usernum, lpc, h;
user* lp;
user* newuser;

	if(strchr(author, ''))
		stripansi(author);

	h = hash(author);

	for (lpc=0, lp = firstuser, usernum=-1; lpc < numuser; ++lpc, lp = lp->next)
		if(h == lp->hash && stricmp(author, lp->username)==0)
		{
			usernum=lpc;
			break;
		}

	if (usernum==-1)
	{
		numuser++;
		usernum=numuser-1;

		if(curruser == NULL) {
			/* no current elements. */
			firstuser = curruser = (user *)malloc(sizeof(user));
			curruser->next = NULL;
		}
		else {
			/* at least one element exists. */
			newuser = (user *)malloc(sizeof(user));
			newuser->next = NULL;
			curruser->next = newuser;
			curruser = newuser;
		}

		if(heapcheck() != 2) {
			printf("(User)The heap is fucked...exiting\n");
			printf("Coreleft   : %u %lu\n", coreleft(), coreleft());
			printf("Farcoreleft: %lu\n", farcoreleft());
			printf("You should probably write to Magrathea @ Skaro about this..\n");
			exit(1);
		}

		strcpy(curruser->username, author);
		curruser->hash = h;
		curruser->messages = 1;
		curruser->msgk = messagebyte;
		curruser->excl = 0;

		/* exclude unwanted nodes */
		for (currxuser = firstxuser; currxuser != NULL; currxuser = currxuser->next)
			if (stricmp(curruser->username, currxuser->name) == 0) {
				if(!status)
					fputc('', stdout);
				curruser->excl = 1;
			}

		if(debug)
			printf("new user: [%25s]\n", author);
		else
		{
			if(status == 0)
				fputc('', stdout);
			else
				if (status == 1) {
					printf("\rUser:%3d. %-30s", numuser, author);
				} else if (status == 2) {
					if (column + 33 > maxcol) {
						printf("\n");
						column = 0;
						}
					column += printf("U:%-30s  ", author);
					}
		}
	}
	else
	{
		lp->messages++;
		lp->msgk += messagebyte;
	}
}

/***************************************************************************/
/* pathstat() - keeps tabs on network connections...                       */
/* ----------------------------------------------------------------------- */
/* I really don't like this function...  I find it quite ugly...           */
/* I'd be delighted if someone could come up with a nicer one...           */
/* I also think it's going to rival userstat() as the most time-intensive  */
/* part of this program...i'll have to profile it again and find out...    */
/***************************************************************************/
void pathstat(void)
{
char	wpath[513];
char	*ptr;			// pointer to start of current part of path
char	*end;			// pointer to ending bang.
char	*point;         // pointer to period in address.
char    add1[31];       // address 1
char	add2[31];		// address 2
char	air[31];		// current airport code.
char	*good;			// and this is a hack.

	strcpy(wpath, path);
	strcat(wpath, homeadd);

	ptr = wpath;

	if(strchr(ptr, '!'))
		do
		{
			good = end = strchr(ptr, '!');
			if(end)
				*end = 0;

			if((point = strchr(ptr, '.')) != NULL)
				strcpy(air, point);

			strcpy(add1, ptr);
			if(!strchr(add1, '.'))
				strcat(add1, air);
			if(end)
				*end = '!';

			good = ptr = end+1;
			good = end = strchr(ptr, '!');
			if(end)
				*end = 0;

			strcpy(add2, ptr);
			if(!strchr(add2, '.'))
				strcat(add2, air);
			if(end)
				*end = '!';

			if(isaddress(add1) && isaddress(add2) && stricmp(add1, add2))
			{
				strcpy(alias1, add1);
				strcpy(alias2, add2);
				linkstat();
			}
		} while(good);
}

/************************************************************************/
/* linkstat() - this is used by pathstat.                               */
/************************************************************************/
void linkstat(void)
{
int linknum, lpc, h1, h2;
link* lp;
link* newlink;

	h1 = hash(alias1);
	h2 = hash(alias2);

	for (lpc=0, lp = firstlink, linknum=-1; lpc < numlink; ++lpc, lp = lp->next)
		if((h1 == lp->hash1 && h2 == lp->hash2) ||
		   (h1 == lp->hash2 && h2 == lp->hash1))
		{
			linknum=lpc;
			break;
		}

	if (linknum==-1)
	{
		numlink++;
		linknum=numlink-1;

		if(currlink == NULL) {
			/* no current elements. */
			firstlink = currlink = (link *)malloc(sizeof(link));
			currlink->next = NULL;
			}
		else {
			/* at least one element exists. */
			newlink = (link *)malloc(sizeof(link));
			newlink->next = NULL;
			currlink->next = newlink;
			currlink = newlink;
			}

		if(debug)
			if(heapcheck() != 2) {
				printf("(Link)The heap is fucked...exiting\n");
				printf("Coreleft   : %u %lu\n", coreleft(), coreleft());
				printf("Farcoreleft: %lu\n", farcoreleft());
				printf("You should probably write to Magrathea @ Skaro about this..\n");
				exit(1);
			}

		strcpy(currlink->add1, alias1);
		strcpy(currlink->add2, alias2);
		currlink->hash1 = h1;
		currlink->hash2 = h2;
		currlink->messages = 1;
		currlink->msgk = messagebyte;

		if(debug)
			printf("new link: <%s!%s>\n", currlink->add1, currlink->add2);
		else
		{
			if(status == 0)
				fputc('!', stdout);
			else
				if (status == 1) {
					printf("\rlink:%3d. %-15s!%-14s", numlink, alias1, alias2);
				} else if (status == 2) {
					if (column + 33 > maxcol) {
						printf("\n");
						column = 0;
						}
					column += printf("L:%-15s!%-14  ", alias1, alias2);
					}
		}
	}
	else
	{
		lp->messages++;
		lp->msgk += messagebyte;
	}


}

/* -------------------------------------------------------------------- */
/*  isaddress() examines string and returns true if it looks like an    */
/*              alias (xxx.xxxx)                                        */
/* --------------------------------------------------------------------
    94Sep26 22:48:28 From Brent Bottles @ The Anticlimactic Teleservice, 
Seattle In WAN Admin>
"St. Dism" continues to be less than valid as an address.				*/

int isaddress(const char *str)
    {
    int len;
    int dotpos;
    int FoundDot = FALSE;
    int i;
 
    len = strlen(str);
 
    // Don't be fooled by "St. Dism"
    if (strchr(str, ' ') != NULL)
        {
        return (FALSE);
        }
 
    for (i = 0; i < len; i++)
        {
        if (str[i] == '.')
            {
            // we can only have one dot.
            if (FoundDot)
                {
                return (FALSE);
                }
 
            dotpos = i;
            FoundDot = TRUE;
            }
        }
 
    if  (   // 1, 2, or 3 chars, no dots
            ((len > 1) && (len < 4) && !FoundDot)
 
            ||
 
            // 4, 5, 6, 7, or 8 chars, there is a dot, and it is at char
            // pos 2, 3, or 4, and not too close to the end.
            (   (len > 3) && (len < 9) && FoundDot && (dotpos > 1) &&
                (dotpos < 4) && ((len - dotpos) < 6)
            )
        )
        {
        return (TRUE);
        }
 
    return (FALSE);
    } 

/*  
 (That 'Don't be fooled by "St. Dism"' comment is a little out-of-date: 
checking for the position of the dot was added later, and "St. Dism" would fail
that.  However, this check is still good to do, as "poo.po o" would pass the 
dotpos test, but fail the strchr() test.) 

------
i'm cute
*/

/* -------------------------------------------------------------------- */
/*  GetStr()        gets a null-terminated string from a file           */
/* -------------------------------------------------------------------- */
int GetStr(FILE *fil, char *str, int mlen)
{
	int   l = 0;
	int  ch;

	do {
		ch=fgetc(fil);
		if(l < mlen)
			str[l]=ch;
		l++;
	} while(ch && ch != EOF);

	if(l < mlen)
		str[l] = 0;
	else
		str[31] = 0;

	return(l);
}

/* -------------------------------------------------------------------- */
/*  LenStr()        Same as GetStr, but this just dumps the input.      */
/*  This is, of course, another attempt at optimization..doesn't really */
/*  help, though.                                                       */
/* -------------------------------------------------------------------- */
int LenStr(FILE *fil)
{
	int   l = 0;
	int  ch;

	do {
		ch=fgetc(fil);
		l++;
	} while(ch && ch != EOF);

	return(l);
}

/* -------------------------------------------------------------------- */
/*  hash()          Make an int out of their name                       */
/* -------------------------------------------------------------------- */
int hash(const char* s)
{
	int h, shift;
	int sp;

	for (h=shift=0, sp=0; s[sp]; shift=(shift+1)&7, sp++)
	{
		h ^= (toupper(s[sp])) << shift;
	}
	return h;
}

/* ------------------------------------------------------------------------ */
/* changes string being sent, does not return a pointer like deansi         */
/* ------------------------------------------------------------------------ */
void stripansi(char *str)
{
	int i=0;
	int a=0;

	char *pc;
	pc = str;

	while (*pc)
	{
		if (*pc == 1)
			a=1;			   /* Zap Ctrl-A plus following Char */
		else
			if(a)
				a=0;
			else {
				str[i] = *pc;
				i++;
			}
		pc++;
	}
	if(i < LABELSIZE)
		str[i] = 0;
	else
		str[31] = 0;
}

/****************************************************************************/
/* cLength() - returns the uncompressed size of a message in bytes.         */
/****************************************************************************/
int cLength(void)
{
unsigned char c_type; // compression type
unsigned char lsb, msb, b;
unsigned short uc_msglen; // uncompressed message length
int i=0;

	c_type = messagetext[i];
	i++;

	if(c_type != 0x01)	// Compression isn't of type 1!  We don't know how
		return length;	// to handle that...so....

	lsb = messagetext[i];
	i++;

	if (lsb == 0x1b)
	{
		b = messagetext[i];
		i++;

		if(b == 0x01)
			lsb = 0x00;
		else
			if(b == 0x02)
				lsb = 0x1b;
			else
				if(b == 0x03)
					lsb = 0xff;
				else
				{
					printf("\nProblem with cLength(lsb)\n%s\n", messagetext);
					return 0;
				}
	}

	// this code gets past the <= 65.155 MSB bug
	msb = messagetext[i];
	i++;

	if (msb == 0x1b)
	{
		b = messagetext[i];
		i++;

		if (b == 0x01)
			if (length < 300)
				msb = 0x00;
			else
				msb = 0x1b;
			else
				if (b == 0x02)
					msb = 0x1b;
				else
					if(b == 0x03)
						msb = 0xff;
					else
					{
						printf("\nProblem with cLength(lsb)\n%s\n", messagetext);
						return 0;
					}
	}
	uc_msglen = lsb | (msb << 8);

	return uc_msglen;
}

/****************************************************************************/
/* sorting stuff...which I don't really understand...                       */
/****************************************************************************/
int Nsort(node* this, node* that) {
	if (strcmp(this->nodename,that->nodename) < 0)
		return -1;
	else
		if (strcmp(this->nodename,that->nodename) > 0)
			return 1;
		else
			return 0;
}

/***************************************/
/* merge 2 lists under dummy head item */
/***************************************/
node* Nlmerge (node* p, node* q) {
	node* r;
	node head;

    for (r = &head; p && q; ) {
		if ( Nsort(p, q) < 0 ) {
			r = r->next = p;
			p = p->next;
		}
		else {
			r = r->next = q;
			q = q->next;
		}
	}
	r->next = (p ? p : q);
	return head.next;
}

/*********************************************************/
/* split list into 2 parts, sort each recursively, merge */
/*********************************************************/
node* Nlsort (node* p) {
	node* q;
	node* r;

    if ( p ) {
        q = p;
        for ( r = q->next; r && (r = r->next) != NULL; r = r->next )
            q = q->next;
        r = q->next;
        q->next = NULL;
        if ( r )
			p = Nlmerge(Nlsort(r), Nlsort(p));
    }
    return p;
}
/***********************/
int Usort(user* this, user* that) {
	if (this->msgk > that->msgk)
		return -1;
	else if (this->msgk < that->msgk)
		return 1;
	else return 0;
}

/***************************************/
/* merge 2 lists under dummy head item */
/***************************************/
user* Ulmerge (user* p, user* q) {
	user* r;
	user head;

    for (r = &head; p && q; ) {
		if ( Usort(p, q) < 0 ) {
			r = r->next = p;
			p = p->next;
		}
		else {
			r = r->next = q;
			q = q->next;
		}
	}
	r->next = (p ? p : q);
	return head.next;
}

/*********************************************************/
/* split list into 2 parts, sort each recursively, merge */
/*********************************************************/
user* Ulsort (user* p) {
	user* q;
	user* r;

    if ( p ) {
        q = p;
        for ( r = q->next; r && (r = r->next) != NULL; r = r->next )
            q = q->next;
        r = q->next;
        q->next = NULL;
        if ( r )
			p = Ulmerge(Ulsort(r), Ulsort(p));
    }
    return p;
}
/****************************/
int Rsort(room* this, room* that) {
	if (this->msgk > that->msgk)
		return -1;
	else if (this->msgk < that->msgk)
		return 1;
	else return 0;
}

/***************************************/
/* merge 2 lists under dummy head item */
/***************************************/
room* Rlmerge (room* p, room* q) {
	room* r;
	room head;

    for (r = &head; p && q; ) {
		if ( Rsort(p, q) < 0 ) {
			r = r->next = p;
			p = p->next;
		}
		else {
			r = r->next = q;
			q = q->next;
		}
	}
	r->next = (p ? p : q);
	return head.next;
}

/*********************************************************/
/* split list into 2 parts, sort each recursively, merge */
/*********************************************************/
room* Rlsort (room* p) {
	room* q;
	room* r;

    if ( p ) {
        q = p;
        for ( r = q->next; r && (r = r->next) != NULL; r = r->next )
            q = q->next;
        r = q->next;
        q->next = NULL;
        if ( r )
			p = Rlmerge(Rlsort(r), Rlsort(p));
    }
    return p;
}

/****************************/
int Lsort(link* this, link* that) {
	if (this->msgk > that->msgk)
		return -1;
	else if (this->msgk < that->msgk)
		return 1;
	else return 0;
}

/***************************************/
/* merge 2 lists under dummy head item */
/***************************************/
link* Llmerge (link* p, link* q) {
	link* r;
	link head;

    for (r = &head; p && q; ) {
		if ( Lsort(p, q) < 0 ) {
			r = r->next = p;
			p = p->next;
		}
		else {
			r = r->next = q;
			q = q->next;
		}
	}
	r->next = (p ? p : q);
	return head.next;
}

/*********************************************************/
/* split list into 2 parts, sort each recursively, merge */
/*********************************************************/
link* Llsort (link* p) {
	link* q;
	link* r;

    if ( p ) {
        q = p;
        for ( r = q->next; r && (r = r->next) != NULL; r = r->next )
            q = q->next;
        r = q->next;
        q->next = NULL;
        if ( r )
			p = Llmerge(Llsort(r), Llsort(p));
    }
    return p;
}

/* -------------------------------------------------------------------- */
/*  startup...  opens all files.                                        */
/* -------------------------------------------------------------------- */
void startup(void)
{
	if((msgdat = fopen(inputmail, "rb")) == NULL) {
		printf(" I can't seem to find \"%s\".\n", inputmail);
		exit(1);
	}

	setvbuf(msgdat, NULL, _IOFBF, 4096);
	setvbuf(stdout, NULL, _IOFBF, 8);	// So I have a slow video card... :-)

	if((nodesout = fopen(outputnode, "rb")) != NULL) {
		printf(" I need to be able to write to \"%s\".\n"
			   " However, it seems to already exist.  You should probably delete it.\n", outputnode);
		exit(1);
	}

#ifdef DEBUG
	if((debugout = fopen(debugfile, "wt")) == NULL) {
		printf(" I need to be able to write to \"%s\".\n"
			   " However, it seems to already exist.  You should probably delete it.\n", debugfile);
		exit(1);
	}
#endif

	if((nodesout = fopen(outputnode, "wt")) == NULL) {
		printf(" I can't seem to open \"%s\" for output.\n", outputnode);
		exit(1);
	}

	fprintf(nodesout, "\n	3List of current active citnet nodes and addresses.0\n"
					  "\n"
					  "This list is created automaticly by scanning the %s message\n"
					  "base for node names.  Thus it is possable for new or quiet nodes to be missed.\n"
					  "\n", homenode);
	if(!jaydee)
	fprintf(nodesout,"Only nodes with more than 8 messages are listed.\n"
					 "Also, nodes running /064 must post more than 57 messages to be included.\n"
					 "\n");

	fprintf(nodesout, "%% msgs is the percentage of total messages that come from the node.\n"
					  "Msgb is the total ammount of message text to come from the node, in bytes.\n"
					  "AVb/msg is the average message size from the node, in bytes.\n"
					  "%% msgb is the percentage of total message text that comes from the node.\n"
					  "\n");

	if((roomsout = fopen(outputroom, "rb")) != NULL) {
		printf(" I need to be able to write to \"%s\".\n"
			   " However, it seems to already exist.  You should probably delete it.\n", outputroom);
		exit(1);
	}

	if((roomsout = fopen(outputroom, "wt")) == NULL) {
		printf(" I can't seem to open \"%s\" for output.\n", outputroom);
		exit(1);
	}

   fprintf(roomsout, "	3List of currently active Citnet rooms.0\n"
					 "\n"
					 "%% msgs is the percentage of total messages that are in that room.\n"
					 "Msgb is the total ammount of message text in the room, in bytes.\n"
					 "AVb/msg is the room's average message size, also in bytes.\n"
					 "%% msgb is the percentage of total network message text in that room.\n"
					 "\n");

   if((usersout = fopen(outputuser, "rb")) != NULL) {
		printf(" I need to be able to write to \"%s\".\n"
			   " However, it seems to already exist.  You should probably delete it.\n", outputuser);
		exit(1);
   }

   if((usersout = fopen(outputuser, "wt")) == NULL) {
		printf(" I can't seem to open \"%s\" for output.\n", outputuser);
		exit(1);
	}

	fprintf(usersout, "	3List of currently active Citnet users.0\n"
					  "\n"
					  "%%msgs is based on the number user messages / total citnet msgs.\n"
					  "%%msgb is based on the total bytes put out by the user / citnet bytes.\n"
					  "\n");

	if((nodeslst = fopen(outnodelist, "rb")) != NULL) {
		printf(" I need to be able to write to \"%s\".\n"
			   " However, it seems to already exist.  You should probably delete it.\n", outnodelist);
	exit(1);
	}

	if((nodeslst = fopen(outnodelist, "wt")) == NULL) {
		printf(" I can't seem to open \"%s\" for output.\n", outnodelist);
		exit(1);
	}

	fprintf(nodeslst, "	3List of current Citnet nodes.0\n"
					  "\n"
					  "Complete with phone numbers and other useful bits...\n"
					  "\n"
					  "Nodename3,0 Region3,0 County3,0 Country 3[0address3]0\n"
					  "Software3,0 Phone number.\n"
					  "\n"
					  "\n");

	if((linkslst = fopen(outlinklist, "rb")) != NULL) {
		printf(" I need to be able to write to \"%s\".\n"
			   " However, it seems to already exist.  You should probably delete it.\n", outlinklist);
	exit(1);
	}

	if((linkslst = fopen(outlinklist, "wt")) == NULL) {
		printf(" I can't seem to open \"%s\" for output.\n", outlinklist);
		exit(1);
	}

	fprintf(linkslst, "	3The Late Show Link List.0\n"
					  "\n"
					  "This is how the network looks from %s's viewpoint.\n"
					  "\n", homenode);
}

/****************************************************************************/
/* output() -- pretty obvious...                                            */
/****************************************************************************/
void output(void)
{
int i;
label tmp;

	fflush(stdout);
	setvbuf(stdout, NULL, _IONBF, 8);

	printf("\nSorting Nodes...");
	firstnode = Nlsort(firstnode);
	printf("Rooms...");
	firstroom = Rlsort(firstroom);
	printf("Users...");
	firstuser = Ulsort(firstuser);
	printf("Links...");
	firstlink = Llsort(firstlink);
	printf("\n");

	fprintf(nodesout, "Nodename                    address   soft  msgs  msgb  AVb/msg %% msgs %% msgb\n");
	fprintf(nodesout, "3a       0\n");

	if(!jaydee)	/* exclude jaydee nodes. */
	{
		for(i=0, currnode = firstnode; currnode != NULL; currnode = currnode->next)
			if(currnode->nodename[0] != '\0' && currnode->messages >= 8 &&
			   !(currnode->messages <= 57 && currnode->software == 1))
				fprintf(nodesout, "%-25.25s (%8.8s) %5s %5ld %6ld %7.2f %5.2f%% %5.2f%%\n",
						currnode->nodename,
						currnode->address,
						softkey[currnode->software],
						currnode->messages,
						currnode->msgk,
						((float)currnode->msgk/currnode->messages),
						((float)currnode->messages/nmessages*100),
						((float)currnode->msgk/messagek*100));
	}
	else	/* do not exclude jaydee nodes. */
	{
		for(i=0, currnode = firstnode; currnode != NULL; currnode = currnode->next)
			fprintf(nodesout, "%-25.25s (%8.8s) %5s %5ld %6ld %7.2f %5.2f%% %5.2f%%\n",
					currnode->nodename,
					currnode->address,
					softkey[currnode->software],
					currnode->messages,
					currnode->msgk,
					((float)currnode->msgk/currnode->messages),
					((float)currnode->messages/nmessages*100),
					((float)currnode->msgk/messagek*100));
	}

	fprintf(nodesout, "\n\tTotal of %d nodes\n", numnode);
	fprintf(nodesout, "\t%d local messages.\n", local);
	fprintf(nodesout, "\tList generated from %ld messages totaling %ld bytes.\n"
					  "\t%ld msg.dat.\n",
			nmessages, messagek, filelength(fileno(msgdat)));

	fprintf(roomsout, "Roomname                        msgs  msgb  AVb/msg %% msgs %% msgb\n");
	fprintf(roomsout, "3a     0\n");

	for(i=0, currroom = firstroom; currroom != NULL; currroom = currroom->next)
		fprintf(roomsout, "%-30.30s %5ld %6ld %7.2f %5.2f%% %5.2f%%\n",
				currroom->roomname,
				currroom->messages,
				currroom->msgk,
				((float)currroom->msgk/currroom->messages),
				((float)currroom->messages/nmessages*100),
				((float)currroom->msgk/messagek*100));

	fprintf(roomsout, "\n\tTotal of %d rooms\n", numroom);
	fprintf(roomsout, "\tList generated from %ld messages totaling %ld bytes.\n"
					  "\t%ld msg.dat.\n",
			nmessages, messagek, filelength(fileno(msgdat)));

	fprintf(usersout, "     Username                        msgs  msgb  AVb/msg %% msgs %% msgb\n");
	fprintf(usersout, "     3a     0\n");

	for(i=0, curruser = firstuser; curruser != NULL;curruser = curruser->next, i++)
		if(curruser->username[0] != 0 && curruser->excl != 1)
			fprintf(usersout, "%3d. %-30.30s %5ld %6lu %7.2f %5.2f%% %5.2f%%\n",
					i+1,
					curruser->username,
					curruser->messages,
					curruser->msgk,
					((float)curruser->msgk/(float)curruser->messages),
					((float)curruser->messages/(float)nmessages*100),
					((float)curruser->msgk/(float)messagek*100));
		else if(debug)
			fprintf(usersout, "Excluded %s\n", curruser->username);


	fprintf(usersout, "\n\tTotal of %d users\n", numuser);
	fprintf(usersout, "\tList generated from %ld messages totaling %ld bytes.\n"
					  "\t%ld msg.dat.\n",
			nmessages, messagek, filelength(fileno(msgdat)));

	for(i=0, currnode = firstnode; currnode != NULL; currnode = currnode->next)
//		if(currnode->nodename[0] != '\0' && currnode->messages >= 8 &&
//			!(currnode->messages <= 57 && currnode->software == 1))
				fprintf(nodeslst, "%s3,0 %s3,0 %s3,0 %s 3[0%s3]0\n"
								  "%s3,0 %s\n"
								  "\n",
								  currnode->nodename,
								  currnode->region,
								  currnode->state,
								  currnode->country,
								  currnode->address,
								  software[currnode->software],
								  currnode->phone);

		fprintf(linkslst, "Link                  msgs   msgb   AVb/msg %% msgs %% msgb\n");
	fprintf(linkslst, "3a     0\n");
		for(i=0, currlink = firstlink; currlink != NULL; currlink = currlink->next)
		{
						sprintf(tmp, "%-8s<->%8s", currlink->add1, currlink->add2);
						fprintf(linkslst, "%-20.20s %6lu %7lu %7.2f %5.2f%% %5.2f%%\n",
								  tmp,
								  currlink->messages,
								  currlink->msgk,
								  ((float)currlink->msgk/currlink->messages),
								  ((float)currlink->messages/nmessages*100),
								  ((float)currlink->msgk/messagek*100));
		}
		fprintf(linkslst, "\n\tTotal of %d network links.\n"
						  "\tList generated from %ld messages totaling %ld bytes.\n"
						  "\t%ld msg.dat.\n",
						  numlink, nmessages, messagek, filelength(fileno(msgdat)));
}

/* from snippets 95/10
   public domain by david engel.
   slightly modified by GS.
*/
FILE *pfopen(const char* name, const char* mode, const char* dirs)
{
    char* ptr;
    char* tdirs;
    FILE* file = NULL;

    if (dirs == NULL || dirs[0] == '\0')
        return NULL;

    if ((tdirs = malloc(strlen(dirs)+1)) == NULL)
        return NULL;

    strcpy(tdirs, dirs);

    for (ptr = strtok(tdirs, ";"); file == NULL && ptr != NULL;
        ptr = strtok(NULL, ";"))
    {
        size_t len;
        char work[FILENAME_MAX];

        strcpy(work, ptr);
        len = strlen(work);
        if (len && work[len-1] != '\\')
            strcat(work, "\\");
        strcat(work, name);

        file = fopen(work, mode);
    }

    free(tdirs);

    return file;
}

/* ------------------------------------------------------------------- */
/* config()  --  Reads a config file.                                  */
/* Lifted by GS from Audin's poster 1.6 source. Not the best config    */
/*  file routines, but they'll work for now!                           */
/* ------------------------------------------------------------------- */
void config(void)
{
    int i;
    FILE *cfg;
    char *text;

    /* we're not in a multitasker, so save stack by making them static */
    static char line[257], word1[257], lword1[257];
    static char drive[5], dir[257], dirlist[257];

    static char *KEYWORD[] = {
        "msgfile",                  /* 0 - msg.dat file */
		"nodelist",                  /* 1 - output file */
        "listtype",                 /* 2 - list type */
        "exclude",                  /* 3 - user exclusion */
		"output",                   /* 4 - output type */
		"userlist", 				// 5 - userlist out
		"roomlist",					// 6 - roomlist
		"nodelist2",                // 7 - extended nodelist
		"nodename",                 // 8 - Host nodename
		"address",                  // 9 - Host address
		"software",                 // 10 - host software
		"region",                   // 11 - host region
		"state",                    // 12 - host state
		"country",                  // 13 - host country
		"phone",                    // 14 - host phone
		"jaydee",					// 15 - exclude jaydee nodes?
		"debug",					// 16 - debug mode?
		"linklist",					// 17 - linklist out.
		NULL
    };

    drive[0] = dir[0] = dirlist[0] = 0;

    /* search these areas for the configuration file:
        current directory
        executable directory
        path
    */
	fnsplit(softname, drive, dir, NULL, NULL);
    sprintf(dirlist, ".;%s%s;%s", drive, dir, getenv("PATH"));
	cfg = pfopen("nodes.cfg", "rt", dirlist);
    if (cfg == NULL) {
        printf("No config file present.\n");
        exit(1);
    }

    printf("Reading config file.\n");

    while (!feof(cfg)) {
        text = fgets(line, sizeof(line), cfg);
        if(text == 0)
            break;
        if(line[0] != '#')
            continue;

        i = sscanf(line, "%s \"%255[^\"\n\r\t]", word1, lword1);

        if (i == 1)
            i = sscanf(line, "%s %255s", word1, lword1);
        if (i != 2)
            continue;

        for (i = 0; KEYWORD[i] != NULL; i++)
            if (stricmp(&word1[1], KEYWORD[i]) == 0)
                break;

        switch (i) {
            case 0:                     /* #msgfile */
                strcpy(inputmail, lword1);
				break;

            case 1:                     /* #outfile */
				strcpy(outputnode, lword1);
				break;

            case 2:                     /* #listtype */
                listtype = atoi(lword1);
				printf("List type: %d\n", listtype);
				break;

            case 3:                     /* #exclude */
                {
                    xuser* newxuser;
                    lword1[31] = 0;
                    if (currxuser == NULL) {
                        /* list has no elements so far */
                        firstxuser = currxuser = (xuser *)malloc(sizeof (xuser));
                        currxuser->next = NULL;
                    }
                    else {
                        /* list has at least one element */
                        newxuser = (xuser *)malloc(sizeof (xuser));
                        newxuser->next = NULL;
                        currxuser->next = newxuser;
                        currxuser = newxuser;
                    }
                    strcpy(currxuser->name, lword1);

						printf("Exclude user: %s\n", currxuser->name);
				}
				break;

            case 4:                     /* #user */
				status = atoi(lword1);
				break;

			case 5:					// userlist out
				strcpy(outputuser, lword1);
				break;

			case 6:					// roomlist out
				strcpy(outputroom, lword1);
				break;

			case 7:					// extended node out
				strcpy(outnodelist, lword1);
				break;

			case 8:					// nodename
				lword1[31] = 0;
				strcpy(homenode, lword1);
				break;

			case 9:					// address
				lword1[31] = 0;
				strcpy(homeadd, "!");
				strcat(homeadd, lword1);
				break;

			case 10:				// software
				lword1[31] = 0;
				strcpy(homesoft, lword1);
				break;

			case 11:				// region
				lword1[31] = 0;
				strcpy(homeregion, lword1);
				break;

			case 12:				// state
				lword1[31] = 0;
				strcpy(homestate, lword1);
				break;

			case 13:				// country
				lword1[31] = 0;
				strcpy(homecountry, lword1);
				break;

			case 14:				// phone
				lword1[31] = 0;
				strcpy(homephone, lword1);
				break;

			case 15:					// jaydee
				if(lword1[0] == '1')
					jaydee = 1;
				break;

			case 16:
				if(lword1[0] == '1')
					debug = 1;
				break;

			case 17:					// linklist out
				strcpy(outlinklist, lword1);
				break;

            default:
                printf("Keyword '%s' unknown.\n", word1);
                exit(1);
                break;
        }
    }

    fclose(cfg);
}
